import uuid
import threading
from typing import List
from .utils import (
    project_explorations_db,
    start_exploration_real_combination_computation,
    get_exploration_by_id,
    update_explorations,
    update_exploration as update_exploration_db,
    create_selection,
    computation_threads,
    stop_flags,
)


def get_exploration_available_config():
    pass


def get_explorations(project_id):
    explorations: List[dict] = project_explorations_db.get(project_id) or []
    for exploration in explorations:
        exploration.pop("combinations", None)
        exploration.pop("metrics", None)
    return explorations


def create_exploration(project_id, body):
    # Get current explorations
    explorations = project_explorations_db.get(project_id) or []

    # Add new exploration
    explorations.append(
        {
            "id": str(uuid.uuid4()),
            "name": body.get("name"),
            "description": body.get("description", ""),
            "state": "not_started",
            "config": {},
        }
    )

    # Update the database
    update_explorations(project_id, explorations)


def get_exploration(project_id, exploration_id):
    return get_exploration_by_id(project_id, exploration_id)


def delete_exploration(project_id, exploration_id):
    # Get current explorations
    explorations = project_explorations_db.get(project_id) or []

    # Filter out the exploration to delete
    explorations = [
        exploration
        for exploration in explorations
        if exploration.get("id") != exploration_id
    ]

    # Update the database
    update_explorations(project_id, explorations)
    return explorations


def update_exploration(project_id, exploration_id, action, body):
    exploration = get_exploration(project_id, exploration_id)
    if exploration is None:
        print("Exploration not found:", exploration_id)
        return None

    # Perform action based on the request
    if action == "start":
        # Check if the exploration is already running
        if exploration.get("state") == "ongoing":
            print("Exploration is already running:", exploration_id)
            return exploration

        # Start the exploration computation in a separate thread
        thread = threading.Thread(
            target=start_exploration_real_combination_computation,
            args=(project_id, exploration_id),
            daemon=True,
        )
        thread.start()

        # Save the thread in the global dictionary
        computation_threads[exploration_id] = thread

        return exploration
    elif action == "stop":
        # Stop the exploration computation
        if exploration_id in computation_threads:
            # Set the stop flag
            stop_flags[exploration_id] = True

            # Wait for the thread to finish
            computation_threads[exploration_id].join()
            del computation_threads[exploration_id]
            del stop_flags[exploration_id]
        elif exploration.get("state") == "ongoing":
            # Exploration is ongoing but no thread found, set state to not_started
            exploration["state"] = "not_started"

        return exploration
    elif action == "updateConfig":
        # Do not update the exploration if it is running
        if exploration.get("state") == "ongoing":
            print("Cannot update config while exploration is running:", exploration_id)
            return exploration

        # Update exploration config
        exploration["config"] = body
        update_exploration_db(project_id, exploration)
        return exploration
    else:
        raise ValueError(f"Unknown action: {action}")


def update_exploration_config():
    pass


def create_exploration_selection(project_id, exploration_id, body):
    # Create a selection for the exploration
    create_selection(
        project_id,
        exploration_id,
        body["selected_combinations"],
        body["selection_name"],
    )

    return 201



#############################################################################
# Imports
#############################################################################
from debiaiServer.modules.dataProviders.DataProviderException import (
    DataProviderException,
)
import debiaiServer.modules.dataProviders.dataProviderManager as data_provider_manager
from debiaiServer.api.v1.debiai.utils import make_hash

#############################################################################
# PROJECTS Management
#############################################################################


def change_project_overview_v1(project_info):
    v1_project_info = {
        "id": project_info["id"],
        "dataProviderProjectId": project_info["id"],  # same id for test
        "name": project_info["name"],
        "creationDate": project_info["creationDate"],
        "updateDate": project_info["updateDate"],
        "tags": [],
        "metadata": {},
        "metrics": {
            "nbModels": project_info["nbModels"],
            "nbSelections": project_info["nbSelections"],
            "nbSamples": project_info["nbSamples"],
        },
        # "columns": column_info,
        # projectColumns, get from statistic et suppress
        # "blockLevelInfo": projectBlockLevel, suppress here keep
    }
    return v1_project_info


def get_projects(prev_hash_content=None):
    # Return a list of project overviews from all the data providers
    data_providers_list = data_provider_manager.get_data_provider_list()
    projectOverviews = {}
    projectList = []
    print("data_provider Get projects")
    for data_provider in data_providers_list:
        print(data_provider)
        try:
            projects = data_provider.get_projects()

            if projects is not None:
                # Adding data provider id to projects
                for project in projects:
                    project_inf = change_project_overview_v1(project)
                    projectList.append(project_inf)

                    if data_provider.name != "Python module Data Provider":
                        project_inf["dataProviderId"] = data_provider.name
                    else:
                        project_inf["dataProviderId"] = "json_block"

        except DataProviderException as e:
            print("Warning get DP projects : " + e.message)

    new_hash = "prj_" + str(
        +make_hash(projectList)
    )  # We add a prefix to avoir empty string

    print(
        new_hash,
        " <=> ",
        prev_hash_content,
        type(new_hash),
        type(prev_hash_content),
        new_hash == prev_hash_content,
    )
    if new_hash == prev_hash_content:
        return None, 304
    else:
        projectOverviews["projects"] = projectList
        projectOverviews["hash_content"] = new_hash
        return projectOverviews, 200



from configparser import ConfigParser
from termcolor import colored
import pkg_resources
import os

config_path = pkg_resources.resource_filename("debiaiServer", "config/config.ini")
config_parser = ConfigParser()

DEBUG_COLOR = "light_blue"
DEBUG_SECONDARY_COLOR = "blue"
ERROR_COLOR = "light_red"
SUCCESS_COLOR = "green"

# Default config
DATA_FOLDER_PATH = "debiai_data"  # The path to the DebiAI data config folder

config = {
    "DATA_PROVIDERS_CONFIG": {
        "creation": True,
        "deletion": True,
    },
    "INTEGRATED_DATA_PROVIDER": {
        "enabled": True,
        "allow_create_projects": True,
        "allow_delete_projects": True,
        "allow_insert_data": True,
        "allow_create_selections": True,
        "allow_delete_selections": True,
        "allow_create_models": True,
        "allow_delete_models": True,
        "allow_insert_results": True,
    },
    "WEB_DATA_PROVIDERS_CONFIG": {
        "cache": True,
        "cache_duration": 120,
    },
    "WEB_DATA_PROVIDERS": {
        # "name": "url"
    },
    "ALGO_PROVIDERS_CONFIG": {
        "enable_integrated": True,
        "creation": True,
        "deletion": True,
    },
    "ALGO_PROVIDERS": {
        # "name": "url"
    },
    "EXPORT_METHODS_CONFIG": {
        "creation": True,
        "deletion": True,
    },
    "EXPORT_METHODS": {
        # "name": "type, param1, param2, ..."
    },
}

# Env vars mapping
ENV_VAR_MAPPING = {
    "DATA_PROVIDERS_CONFIG": {
        "creation": "DEBIAI_DATA_PROVIDERS_ALLOW_CREATION",
        "deletion": "DEBIAI_DATA_PROVIDERS_ALLOW_DELETION",
    },
    "INTEGRATED_DATA_PROVIDER": {
        "enabled": "DEBIAI_INTEGRATED_DATA_PROVIDER_ENABLED",
        "allow_create_projects": "DEBIAI_INTEGRATED_DP_ALLOW_CREATE_PROJECTS",
        "allow_delete_projects": "DEBIAI_INTEGRATED_DP_ALLOW_DELETE_PROJECTS",
        "allow_insert_data": "DEBIAI_INTEGRATED_DP_ALLOW_INSERT_DATA",
        "allow_create_selections": "DEBIAI_INTEGRATED_DP_ALLOW_CREATE_SELECTIONS",
        "allow_delete_selections": "DEBIAI_INTEGRATED_DP_ALLOW_DELETE_SELECTIONS",
        "allow_create_models": "DEBIAI_INTEGRATED_DP_ALLOW_CREATE_MODELS",
        "allow_delete_models": "DEBIAI_INTEGRATED_DP_ALLOW_DELETE_MODELS",
        "allow_insert_results": "DEBIAI_INTEGRATED_DP_ALLOW_INSERT_RESULTS",
    },
    "WEB_DATA_PROVIDERS_CONFIG": {
        "cache": "DEBIAI_WEB_DATA_PROVIDERS_CACHE_ENABLED",
        "cache_duration": "DEBIAI_WEB_DATA_PROVIDERS_CACHE_DURATION",
    },
    "ALGO_PROVIDERS_CONFIG": {
        "enable_integrated": "DEBIAI_ALGO_PROVIDERS_ENABLE_INTEGRATED",
        "creation": "DEBIAI_ALGO_PROVIDERS_ALLOW_CREATION",
        "deletion": "DEBIAI_ALGO_PROVIDERS_ALLOW_DELETION",
    },
    "EXPORT_METHODS_CONFIG": {
        "creation": "DEBIAI_EXPORT_METHODS_ALLOW_CREATION",
        "deletion": "DEBIAI_EXPORT_METHODS_ALLOW_DELETION",
    },
}

# List of list based config sections + their env var mapping
LIST_CONFIG_SECTIONS = {
    "WEB_DATA_PROVIDERS": "DEBIAI_WEB_DATA_PROVIDER_",
    "ALGO_PROVIDERS": "DEBIAI_ALGO_PROVIDER_",
    "EXPORT_METHODS": "DEBIAI_EXPORT_METHOD_",
}

changes_made = False


def get_config_value(section, key, parameters, config_parser):
    # Config priority order: parameters > config file > env var > default value

    # Get the value from the parameters
    if key in parameters:
        return str.lower(parameters[key])

    # Get the value from the environment variables
    ENV_VAR = ENV_VAR_MAPPING[section][key]
    if ENV_VAR in os.environ:
        return str.lower(os.environ[ENV_VAR])

    # Get the value from the config file
    if section in config_parser and key in config_parser[section]:
        return str.lower(config_parser[section][key])

    print(
        " - Missing "
        + colored(section, DEBUG_SECONDARY_COLOR)
        + " / "
        + colored(key, DEBUG_SECONDARY_COLOR)
        + " in config or in "
        + colored(ENV_VAR, DEBUG_SECONDARY_COLOR)
        + " env var, using default"
    )


def get_config_values(section, parameters, config_parser):
    # Config priority order: parameters > config file > env var > default value

    values = {}

    # Get the value from the config file
    if section in config_parser:
        for key in config_parser[section]:
            values[key] = str.lower(config_parser[section][key])

    # Get the value from the environment variables
    # iterate over the keys of the env var
    ENV_VAR = LIST_CONFIG_SECTIONS[section]
    for key in os.environ.keys():
        if key.startswith(ENV_VAR):
            # Get the key name without the env var prefix
            key_name = key[len(ENV_VAR) :]  # noqa
            values[key_name] = str.lower(os.environ[key])

    # Get the value from the parameters
    for key in parameters:
        if key.startswith(ENV_VAR):
            # Get the key name without the env var prefix
            key_name = key[len(ENV_VAR) :]  # noqa
            values[key_name] = str.lower(parameters[key])

    return values


def set_config_value(section, key, value):
    global changes_made

    if section in config and key in config[section]:
        if config[section][key] != value:
            # The default value is different from the one in the config file
            config[section][key] = value
            changes_made = True

            print(
                " - Overriding "
                + colored(section, DEBUG_COLOR)
                + " / "
                + colored(key, DEBUG_COLOR)
                + " with value "
                + colored(str(value), DEBUG_COLOR)
            )


def init_config(data_folder_path: str = None, parameters: dict = {}):
    global DATA_FOLDER_PATH

    print("===================== CONFIG =======================")

    # Apply the given data_folder_path
    if data_folder_path:
        DATA_FOLDER_PATH = data_folder_path

    # Read the config file
    config_parser.read(config_path)

    for section in config.keys():
        # Deal with boolean, integer and string values
        for key in config[section].keys():
            # Get the value from the config file or the environment variables
            value = get_config_value(section, key, parameters, config_parser)

            if value is None:
                continue

            # Deal with booleans
            if type(config[section][key]) is bool:
                if value == "false":
                    set_config_value(section, key, False)
                elif value == "true":
                    set_config_value(section, key, True)
                else:
                    print(
                        colored("   [ERROR]", ERROR_COLOR)
                        + " Invalid boolean value for "
                        + colored(key, DEBUG_COLOR)
                        + ", using default value"
                    )
                    continue

            # Deal with integers
            elif type(config[section][key]) is int:
                try:
                    set_config_value(section, key, int(value))
                except ValueError:
                    print(
                        colored("   [ERROR]", ERROR_COLOR)
                        + " Invalid integer value for "
                        + colored(key, DEBUG_COLOR)
                        + ", using default value"
                    )
                    continue

            # Deal with strings
            elif type(config[section][key]) is str:
                set_config_value(section, key, str(value))

        # Deal with list based config elements
        if section in LIST_CONFIG_SECTIONS:
            elements = get_config_values(section, parameters, config_parser)

            for element_name in elements:
                print(
                    " - Adding "
                    + section.lower().replace("_", "-")[0:-1]
                    + " "
                    + colored(element_name, DEBUG_COLOR)
                    + " ("
                    + colored(elements[element_name], DEBUG_COLOR)
                    + ")"
                )

                config[section][element_name] = elements[element_name]

    if not changes_made:
        print("   Default config used")


def get_data_folder_path():
    return DATA_FOLDER_PATH


def get_config():
    return config



#############################################################################
# Imports
#############################################################################
import debiaiServer.utils.widgetConfigurations.widgetConfigurations as widgetConfUtils

#############################################################################
# Widget configuration Management
#############################################################################


def get_all_configurations():
    configurations_overview = widgetConfUtils.get_configurations_overview()
    return configurations_overview, 200


def get_widget_configurations(widgetKey):
    configurations = widgetConfUtils.get_configurations(widgetKey)
    return configurations, 200


def post_configuration(widgetKey, data):
    widgetConfUtils.add_configuration(widgetKey, data)
    return None, 204


def delete_configuration(widgetKey, id):
    widgetConfUtils.delete_configuration(widgetKey, id)
    return None, 204



#############################################################################
# Imports
#############################################################################
from debiaiServer.modules.dataProviders.DataProviderException import (
    DataProviderException,
)
import debiaiServer.modules.dataProviders.dataProviderManager as data_provider_manager
import debiaiServer.modules.exportMethods.exportUtils as exportUtils
from debiaiServer.api.v1.exploration_statistics.utils import get_columns_statistics

#############################################################################
# Internal Data Provider
#############################################################################

dataProviderId = "json_block"


def get_data_providers_project():

    # Return a list of project overviews for a specific data provider
    data_provider = data_provider_manager.get_single_data_provider(dataProviderId)

    if data_provider is None:
        return "Data provider not found", 404

    try:
        projects = data_provider.get_projects()

        if projects is not None:
            # Adding data provider id to projects
            for project in projects:
                project["dataProviderId"] = data_provider.name

    except DataProviderException as e:
        print("Warning get DP projects : " + e.message)

    return projects, 200


def change_project_v1(project_info, column_info):
    v1_project_info = {
        "id": project_info["id"],
        "name": project_info["name"],
        "creationDate": project_info["creationDate"],
        "updateDate": project_info["updateDate"],
        "tags": [],
        "metadata": {},
        "metrics": {
            "nbModels": project_info["nbModels"],
            "nbSelections": project_info["nbSelections"],
            "nbSamples": project_info["nbSamples"],
        },
        "columns": column_info,
        # "nbModels": project_info["nbModels"],
        # "nbSelections": project_info["nbSelections"],
        # "nbSamples": project_info["nbSamples"],
        # "blockLevelInfo": project_info["blockLevelInfo"],
        "selections": project_info["selections"],
        "models": project_info["models"],
        # projectColumns, get from statistics and remove duplicates
        # "blockLevelInfo": projectBlockLevel, remove here, keep in the API for python module in V1
    }

    return v1_project_info


def get_project(projectId):
    # return the info about datasets, models, selections & tags
    try:
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)

        project = data_provider.get_project(projectId)

        stats = get_columns_statistics(data_provider.name, project["id"])
        project_inf = change_project_v1(project, stats["columns"])

        # Adding data provider id to project
        project_inf["dataProviderId"] = dataProviderId

        return project_inf, 200
    except DataProviderException as e:
        return e.message, e.status_code


def delete_project(projectId):
    # Delete a project
    try:
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)

        data_provider.delete_project(projectId)
        return "Project deleted", 200
    except DataProviderException as e:
        return e.message, e.status_code


def get_data_id_list(projectId, requestParameters):
    # return the list of data ids
    try:
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)

        data_id_list = data_provider.get_id_list(
            projectId,
            requestParameters["analysis"],
            requestParameters["from"],
            requestParameters["to"],
        )

        return data_id_list, 200
    except DataProviderException as e:
        return e.message, e.status_code


def get_model_id_list(projectId, modelId):
    """
    Get the list of models for a project
    """
    try:
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)
        return list(data_provider.get_model_results_id_list(projectId, modelId)), 200
    except DataProviderException as e:
        return e.message, e.status_code


def delete_model(projectId, modelId):
    """
    Delete a model
    """
    try:
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)
        return data_provider.delete_model(projectId, modelId), 200
    except DataProviderException as e:
        return e.message, e.status_code


def get_results(projectId, modelId, data):
    """
    Get the model results from a sample list
    """
    try:
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)
        return (
            data_provider.get_model_results(projectId, modelId, data["sampleIds"]),
            200,
        )
    except DataProviderException as e:
        return e.message, e.status_code


def get_data(projectId, data):
    # return a project data from a list of ids
    sampleIds = data["sampleIds"]
    analysis = data["analysis"] if "analysis" in data else None

    try:
        # Find the data provider
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)

        # Ask for the data
        samples = data_provider.get_samples(projectId, analysis, sampleIds)

        if samples is not None:
            return {
                "data": samples,
                "dataMap": True,
            }, 200

        return (
            "Can't find samples for project "
            + projectId
            + " on data provider : "
            + dataProviderId,
            404,
        )
    except DataProviderException as e:
        return e.message, e.status_code


def get_selections(projectId):
    try:
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)
        return data_provider.get_selections(projectId), 200
    except DataProviderException as e:
        return e.message, e.status_code


def post_selection(projectId, data):
    try:
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)
        new_selection = data_provider.create_selection(
            projectId,
            data["selectionName"],
            data["sampleHashList"],
        )
        return new_selection, 200
    except DataProviderException as e:
        return e.message, e.status_code


def get_selection_id_list(projectId, selectionId):
    try:
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)
        return data_provider.get_selection_id_list(projectId, selectionId), 200
    except DataProviderException as e:
        return e.message, e.status_code


def delete_selection(projectId, selectionId):
    try:
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)
        data_provider.delete_selection(projectId, selectionId)
        return "Selection deleted", 200
    except DataProviderException as e:
        return e.message, e.status_code


def exportSelection(dataProviderId, projectId, data):
    try:
        return exportUtils.exportSelection(dataProviderId, projectId, data), 200
    except Exception as e:
        return str(e), 400



import os
import ujson as json
from debiaiServer.modules.dataProviders.pythonDataProvider.dataUtils import (
    pythonModuleUtils,
    projects,
    tree,
)
from debiaiServer.modules.dataProviders.DataProviderException import (
    DataProviderException,
)

DATA_PATH = pythonModuleUtils.DATA_PATH


# Models
def get_model_ids(project_id):
    return os.listdir(DATA_PATH() + project_id + "/models/")


def get_models(project_id):
    ret = []
    for model in os.listdir(DATA_PATH() + project_id + "/models/"):
        with open(
            DATA_PATH() + project_id + "/models/" + model + "/info.json"
        ) as json_file:
            info = json.load(json_file)
            ret.append(
                {
                    "name": model,
                    "id": model,
                    "creationDate": info["creationDate"],
                    "updateDate": info["updateDate"],
                    "version": "0.0.0",
                    "metadata": info["metadata"],
                    "nbResults": info["nbResults"],
                }
            )

    return ret


def model_exist(project_id, model_id):
    return model_id in get_model_ids(project_id)


def create_model(project_id, model_name, metadata=None):
    # ParametersCheck
    if not pythonModuleUtils.is_filename_clean(model_name):
        raise DataProviderException("Model name contain invalid characters", 402)

    model_id = model_name

    if model_exist(project_id, model_id):
        raise DataProviderException("Model " + model_id + " already exists", 409)

    if metadata is None:
        metadata = {}

    # model
    modelFolderPath = DATA_PATH() + project_id + "/models/" + model_id
    os.mkdir(modelFolderPath)

    now = pythonModuleUtils.timeNow()

    model_info = {
        "name": model_id,
        "id": model_id,
        "creationDate": now,
        "updateDate": now,
        "metadata": metadata,
        "nbResults": 0,
    }

    pythonModuleUtils.writeJsonFile(modelFolderPath + "/info.json", model_info)

    # Add 0 results to init the file
    write_model_results(project_id, model_id, {})


def delete_model(project_id, model_id):
    pythonModuleUtils.deleteDir(DATA_PATH() + project_id + "/models/" + model_id)


def write_model_results(project_id, model_id, results):
    pythonModuleUtils.writeJsonFile(
        DATA_PATH() + project_id + "/models/" + model_id + "/results.json", results
    )
    projects.update_project(project_id)


def get_model_results(project_id, model_id, sample_ids):
    # Check parameters
    if not projects.project_exist(project_id):
        raise ("Project '" + project_id + "' doesn't exist")
    if not model_exist(project_id, model_id):
        raise ("Model " + model_id + " does not exist")

    # Get model results
    with open(
        DATA_PATH() + project_id + "/models/" + model_id + "/results.json", "r"
    ) as jsonFile:
        model_results = json.load(jsonFile)

    # if not selection_id:
    # return d
    # else:
    # selectionSamples = set(
    #     selections.getSelectionSamples(project_id, selection_id))
    # return selectionSamples.intersection_update(d)
    #     model_results = getModelResults(project_id, model_id)

    ret = {}
    for sample_id in sample_ids:
        if sample_id in model_results:
            ret[sample_id] = model_results[sample_id]
        # Not sending error if sample not found in model results at the moment
        # else:
        #     raise ValueError("Sample " + sample_id +
        #                      " not found in model results")
    return ret


def get_model_id_list(project_id, model_id) -> list:
    # Get model results
    with open(
        DATA_PATH() + project_id + "/models/" + model_id + "/results.json", "r"
    ) as jsonFile:
        model_results = json.load(jsonFile)
        return dict.keys(model_results)


# def get_model_list_results(project_id, model_ids: list, common: bool) -> list:
#     samples = set(get_model_results(project_id, model_ids[0]))

#     for model_id in model_ids[1:]:
#         if common:  # Common samples between models
#             samples.intersection_update(
#                 get_model_results(project_id, model_id))
#         else:  # Union of the model results samples
#             samples = samples.union(get_model_results(project_id, model_id))

#     return list(samples)


def add_results_dict(project_id, modelId, data):
    tree = data["results"]

    # Check parameters
    if not projects.project_exist(project_id):
        raise "Project '" + project_id + "' doesn't exist"

    if not model_exist(project_id, modelId):
        raise (
            "Model '" + modelId + "' in project : '" + project_id + "' doesn't exist"
        )

    # Get resultStructure & project_block_structure
    result_structure = projects.get_result_structure(project_id)
    if result_structure is None:
        raise (
            "The project expected results need to be specified before adding results"
        )

    if "expected_results_order" in data:
        expected_results_order = data["expected_results_order"]
    else:
        expected_results_order = list(map(lambda r: r["name"], result_structure))

    project_block_structure = projects.get_project_block_level_info(project_id)
    sampleIndex = len(project_block_structure) - 1

    # Check the given expected_results_order
    for expected_result in result_structure:
        if expected_result["name"] not in expected_results_order:
            raise (
                "The expected result '"
                + expected_result["name"]
                + "' is missing from the expected_results_order Array"
            )

    giv_exp_res = {}
    for given_expected_result in expected_results_order:
        result_expected = False
        for i, expected_result in enumerate(result_structure):
            if given_expected_result == expected_result["name"]:
                result_expected = True

                # Map the expected_results_order indexes to result_structure
                giv_exp_res[given_expected_result] = i

        if not result_expected:
            return (
                "The given expected result '"
                + given_expected_result
                + "' is not an expected result",
                403,
            )

    #  Check if all blocks referenced in the result tree exists
    resultsToAdd = {}

    for blockKey in tree:
        ok, msg = __check_blocks_of_tree_exists(
            project_id,
            result_structure,
            giv_exp_res,
            tree[blockKey],
            0,
            sampleIndex,
            blockKey,
            resultsToAdd,
        )
        if not ok:
            print(msg)
            return msg, 403

    # The given tree is compliant, let's add the results
    newResults = pythonModuleUtils.addToJsonFIle(
        DATA_PATH() + project_id + "/models/" + modelId + "/results.json", resultsToAdd
    )

    pythonModuleUtils.addToJsonFIle(
        DATA_PATH() + project_id + "/models/" + modelId + "/info.json",
        {"nbResults": len(newResults), "updateDate": pythonModuleUtils.timeNow()},
    )
    projects.update_project(project_id)
    return 200


def __check_blocks_of_tree_exists(
    project_id: str,
    result_structure: list,
    giv_exp_res: dict,
    block: dict,
    level: int,
    sampleIndex: int,
    path: str,
    resultsToAdd: dict,
):
    # Check block exist in the data
    blockInfo = tree.findBlockInfo(project_id, path)
    if not blockInfo:
        return (
            False,
            "Error while adding the results : block '" + path + "' doesn't exist",
        )

    if level == sampleIndex:
        path += "/"
        resultsToAdd[blockInfo["id"]] = []
        #  Sample level : the results : they need to be verified
        if len(block) != len(giv_exp_res):
            raise ValueError(
                "in : "
                + path
                + ", "
                + str(len(block))
                + " value where given but "
                + str(len(giv_exp_res))
                + "where expected"
            )

        for result in result_structure:
            resultsToAdd[blockInfo["id"]].append(block[giv_exp_res[result["name"]]])

        return True, None

    for subBlockKey in block:
        ok, msg = __check_blocks_of_tree_exists(
            project_id,
            result_structure,
            giv_exp_res,
            block[subBlockKey],
            level + 1,
            sampleIndex,
            path + "/" + str(subBlockKey),
            resultsToAdd,
        )
        if not ok:
            return False, msg

    return True, ""



import time
import pkg_resources
from urllib.parse import urlparse
from itertools import cycle
from shutil import get_terminal_size
from threading import Thread
from time import sleep


def get_app_version():
    import yaml
    from yaml.loader import SafeLoader

    # Read the version from the API YAML file
    yaml_path = pkg_resources.resource_filename("debiaiServer", "swagger.yaml")

    try:
        with open(yaml_path, "r") as f:
            data = yaml.load(f, Loader=SafeLoader)
            return data["info"]["version"]
    except Exception as e:
        print(e)
        return "?.?.?"


# Date
def timeNow():
    return time.time() * 1000


# Url
def is_url_valid(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except Exception:
        return False


# Name
def is_valid_name(name):
    # /, &, | are not allowed in data-providers & algo-providers names
    if (
        "/" in name
        or "&" in name
        or "|" in name
        or len(name) == 0
        or len(name) > 50
        or name[0] == " "
        or name[-1] == " "
    ):
        return False

    return True


class Loader:
    def __init__(self, desc="Loading...", end="Done!", timeout=0.1):
        """
        A loader-like context manager

        Args:
            desc (str, optional): The loader's description. Defaults to "Loading...".
            end (str, optional): Final print. Defaults to "Done!".
            timeout (float, optional): Sleep time between prints. Defaults to 0.1.
        """
        self.desc = desc
        self.end = end
        self.timeout = timeout

        self._thread = Thread(target=self._animate, daemon=True)
        self.steps = ["⢿", "⣻", "⣽", "⣾", "⣷", "⣯", "⣟", "⡿"]
        self.done = False

    def start(self):
        self._thread.start()
        return self

    def _animate(self):
        for c in cycle(self.steps):
            if self.done:
                break
            print(f"\r{self.desc} {c}", flush=True, end="")
            sleep(self.timeout)

    def __enter__(self):
        self.start()

    def stop(self):
        self.done = True
        cols = get_terminal_size((80, 20)).columns
        print("\r" + " " * cols, end="", flush=True)
        print(f"\r{self.end}", flush=True)

    def __exit__(self, exc_type, exc_value, tb):
        # handle exceptions with those variables ^
        self.stop()



# This service is used to cache data from the web data provider
# It's used to avoid multiple requests to the web data provider
# It will mainly save the id list of samples, selections and models results
# The ability to cache and the time to live are configurable in the config file

from debiaiServer.config.init_config import get_config
from cacheout import Cache as CacheoutCache


class Cache:
    def __init__(self):
        # Get config
        self.config = get_config()

        self.cache_enabled = self.config["WEB_DATA_PROVIDERS_CONFIG"]["cache"]
        self.cache_ttl = self.config["WEB_DATA_PROVIDERS_CONFIG"]["cache_duration"]

        # Init cache
        self.project_id_list_cache = CacheoutCache(maxsize=256, ttl=self.cache_ttl)
        # <id_project>_<from>_<to>: [...]
        # <id_project>_total: [...]

        self.selection_id_list_cache = CacheoutCache(maxsize=256, ttl=self.cache_ttl)
        # <id_project>_<id_selection>: [...]

        self.model_result_id_list_cache = CacheoutCache(maxsize=256, ttl=self.cache_ttl)
        # <id_project>_<id_model>: [...]

    # Project id list
    def get_key(self, id_project, _from=None, _to=None):
        if _from is None or _to is None:
            return "{}_total".format(id_project)
        else:
            return "{}_{}_{}".format(id_project, _from, _to)

    def get_id_list(self, id_project, _from=None, _to=None):
        if not self.cache_enabled:
            return None

        key = self.get_key(id_project, _from, _to)

        return self.project_id_list_cache.get(key)

    def set_id_list(self, id_project, id_list, _from=None, _to=None):
        if not self.cache_enabled:
            return

        key = self.get_key(id_project, _from, _to)

        self.project_id_list_cache.set(key, id_list)

    # Selection id list
    def get_selection_key(self, id_project, id_selection):
        return "{}_{}".format(id_project, id_selection)

    def get_selection_id_list(self, id_project, id_selection):
        if not self.cache_enabled:
            return None

        key = self.get_selection_key(id_project, id_selection)

        return self.selection_id_list_cache.get(key)

    def set_selection_id_list(self, id_project, id_selection, id_list):
        if not self.cache_enabled:
            return

        key = self.get_selection_key(id_project, id_selection)

        self.selection_id_list_cache.set(key, id_list)

    # Model result id list
    def get_model_result_key(self, id_project, id_model):
        return "{}_{}".format(id_project, id_model)

    def get_model_result_id_list(self, id_project, id_model):
        if not self.cache_enabled:
            return None

        key = self.get_model_result_key(id_project, id_model)

        return self.model_result_id_list_cache.get(key)

    def set_model_result_id_list(self, id_project, id_model, id_list):
        if not self.cache_enabled:
            return

        key = self.get_model_result_key(id_project, id_model)

        self.model_result_id_list_cache.set(key, id_list)



# Utils function use by severals apis


def freeze(o):

    if isinstance(o, dict):
        return frozenset({k: freeze(v) for k, v in o.items()}.items())

    if isinstance(o, list):
        return tuple([freeze(v) for v in o])

    return o


def make_hash(o):
    """
    makes a hash out of anything that contains only list,dict and hashable types including string and numeric types
    """
    return hash(freeze(o))



# Description: Exception class for AlgoProvider
class AlgoProviderException(Exception):
    message = "AlgoProvider error"
    status_code = 500

    def __init__(self, message=None, status_code=None):
        super(AlgoProviderException, self).__init__(message)

        if message is not None:
            self.message = message
        if status_code is not None:
            self.status_code = status_code

    def __str__(self):
        return self.message



import os
from termcolor import colored

from debiaiServer.config.init_config import DEBUG_COLOR
from debiaiServer.modules.algoProviders.AlgoProvider import AlgoProvider
from debiaiServer.modules.algoProviders.AlgoProviderException import (
    AlgoProviderException,
)


def _get_algorithm_python(algorithm_name):
    """Get the python file of the algorithm

    Args:
        algorithm_name (str): Name of the algorithm

    Returns:
        module: Python module of the algorithm
    """

    # Get the algorithm file
    algorithm_file = None
    for file in os.listdir(os.path.dirname(__file__) + "/algorithms"):
        if file.endswith(".py") and file[:-3] == algorithm_name:
            algorithm_file = file[:-3]
            break

    # Check if the file exists
    if algorithm_file is None:
        raise ValueError("Algorithm " + algorithm_name + " does not exists")

    # Import the algorithm
    algorithm_python = __import__(
        "debiaiServer.modules.algoProviders.integratedAlgoProvider.algorithms."
        + algorithm_file,
        fromlist=["*"],
    )

    return (algorithm_name, algorithm_python)


class IntegratedAlgoProvider(AlgoProvider):
    # Integrated AlgoProvider
    # Used to expose the algorithms that are integrated
    # directly in DebiAI
    def __init__(self):
        self.url = "/app/algo-provider"
        self.name = "Integrated Algo-provider"
        self.alive = True

    def is_alive(self):
        return True

    def get_algorithms(self):
        """Get all algorithms that DebiAI can provide
        Returns:
            list: List of algorithms
        """

        # List the .py files if the algorithms folder
        algorithm_files = []
        for file in os.listdir(os.path.dirname(__file__) + "/algorithms"):
            if file.endswith(".py") and file != "__init__.py":
                algorithm_files.append(file[:-3])

        # Import the algorithms
        algorithms_python = []
        for file in algorithm_files:
            print("   Importing " + colored(file, DEBUG_COLOR))
            try:
                algorithms_python.append(_get_algorithm_python(file))
            except ModuleNotFoundError as e:
                print("Error importing " + file)
                print(e)

        # Get the algorithms (call the get_algorithm_details() function)
        algorithms = []
        for algorithm in algorithms_python:
            algorithm_details = algorithm[1].get_algorithm_details()
            # Add the id as the file name
            algorithm_details["id"] = algorithm[0]
            algorithms.append(algorithm_details)

        return algorithms

    def use_algorithm(self, algorithm_id, data):
        try:
            print("Using integrated algo-provider")
            print("Using algorithm: " + algorithm_id)
            algorithm = _get_algorithm_python(algorithm_id)

            # Use the algorithm
            outputs = algorithm[1].use_algorithm(data["inputs"])

            return outputs

        except TypeError as e:
            print("The integrated algo-provider returned an error")
            print(e)
            raise AlgoProviderException(
                algorithm_id + " returned an error: " + str(e), 400
            )
        except Exception as e:
            print("The integrated algo-provider returned an error")
            print(e)
            raise AlgoProviderException("AlgoProvider internal server error", 500)



import requests
import json
from debiaiServer.modules.dataProviders.DataProviderException import (
    DataProviderException,
)


# Todo : change info if in not alive anymore
def get_status(url):
    try:
        r = requests.get(url + "/info")

        if r.status_code != 200:
            return False

        # Check content type
        content = get_http_response(r)

        if content is None:
            return False  # we are expecting a dict

        return True

    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return False
    except requests.exceptions.InvalidURL:
        raise DataProviderException("Invalid URL", 400)


def get_info(url) -> dict:
    try:
        r = requests.get(url + "/info")
        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return {}


# ==== Projects ====
def get_projects(url):
    try:
        r = requests.get(url + "/projects")
        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return None


def get_project(url, id_project):
    try:
        r = requests.get(url + "/projects/" + id_project)
        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return None


def get_id_list(url, id_project, analysis, _from=None, _to=None):
    try:
        if _from is not None and _to is not None:
            url = (
                url
                + "/projects/"
                + id_project
                + "/dataIdList?from={}&to={}&analysisId={}".format(
                    _from, _to, analysis["id"]
                )
            )
        else:
            url = (
                url
                + "/projects/"
                + id_project
                + "/dataIdList?analysisId={}".format(analysis["id"])
            )

        if "start" in analysis and analysis["start"]:
            url += "&analysisStart={}".format(str(analysis["start"]).lower())
        if "end" in analysis and analysis["end"]:
            url += "&analysisEnd={}".format(str(analysis["end"]).lower())

        r = requests.post(url)

        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        print(
            "Error getting data id list from {} on project {}".format(url, id_project)
        )
        return []


def get_samples(url, id_project, analysis, id_list):
    try:
        if analysis:
            rurl = (
                url
                + "/projects/{}/blocksFromSampleIds?analysisId={}&\
analysisStart={}&analysisEnd={}".format(
                    id_project,
                    analysis["id"],
                    str(analysis["start"]).lower() if "start" in analysis else "false",
                    str(analysis["end"]).lower() if "end" in analysis else "false",
                )
            )
        else:
            rurl = url + "/projects/{}/blocksFromSampleIds".format(id_project)

        r = requests.post(rurl, json={"sampleIds": id_list})
        print(r.json())
        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        raise Exception(
            "Could not get the data provider {} data for project {}".format(
                url, id_project
            )
        )


def delete_project(url, id_project):
    try:
        r = requests.delete(url + "/projects/" + id_project)
        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return None


# ==== Selections ====
def get_selections(url, id_project):
    try:
        r = requests.get(url + "/projects/{}/selections".format(id_project))
        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return None


def post_selection(url, id_project, data):
    try:
        r = requests.post(url + "/projects/{}/selections".format(id_project), json=data)
        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return None


def get_selection_id(url, id_project, id_selection):
    try:
        r = requests.get(
            url
            + "/projects/{}/selections/{}/selected-data-id-list".format(
                id_project, id_selection
            )
        )

        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return None


def delete_selection(url, id_project, id_selection):
    try:
        r = requests.delete(
            url + "/projects/{}/selections/{}".format(id_project, id_selection)
        )

        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return None


# ==== Models ====
def get_models(url, id_project):
    try:
        r = requests.get(url + "/projects/{}/models".format(id_project))
        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return None


def get_model_result_id_list(url, project_id, model_id):
    try:
        r = requests.get(
            url
            + "/projects/{}/models/{}/evaluated-data-id-list".format(
                project_id, model_id
            )
        )
        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return None


def get_model_result(url, id_project, id_model, id_sample_list):
    try:
        r = requests.post(
            url + "/projects/{}/models/{}/results".format(id_project, id_model),
            json=id_sample_list,
        )
        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return None


def delete_model(url, id_project, id_model):
    try:
        r = requests.delete(url + "/projects/{}/models/{}".format(id_project, id_model))
        return get_http_response(r)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        return None


# ==== Utils ====
def get_http_response(response):
    try:
        if response.raise_for_status() is None:
            return get_valid_response(response)
    except requests.exceptions.HTTPError:
        return get_error_response(response)


def get_valid_response(response):
    if response.status_code == 204:
        return True
    try:
        return response.json()
    except json.decoder.JSONDecodeError:
        return


def get_error_response(response):
    if response.status_code == 500:
        raise DataProviderException("Data Provider unexpected Error", 500)

    raise DataProviderException(response.text, response.status_code)



from debiaiServer.modules.dataProviders.pythonDataProvider.dataUtils import (
    pythonModuleUtils,
    tree,
    hash,
)

DATA_TYPES = pythonModuleUtils.DATA_TYPES


# ID list
def get_all_samples_id_list(project_id, _from=None, _to=None):
    """
    Return a list of all samples id in a project
    """
    # Get the hashmap
    hashmap = hash.getHashmap(project_id)

    # Get all samples
    samples = list(hashmap.keys())

    # In case of streaming purpose
    if _from is not None and _to is not None:
        samples = samples[_from : _to + 1]  # noqa

    return samples


def get_non_existing_ids(project_id, id_list):
    """
    Return a list of all samples id that do not exist in a project
    """
    # Get the hashmap
    hashmap = hash.getHashmap(project_id)

    # Get all samples
    samples = list(hashmap.keys())

    # Get non existing samples
    non_existing_samples = []
    for sample_id in id_list:
        if sample_id not in samples:
            non_existing_samples.append(sample_id)

        if len(non_existing_samples) > 30:
            break

    return non_existing_samples


# Get data
def get_data_from_sample_id_list(project_id, id_list):
    # Get path of the samples from the hashmap
    sample_path = hash.getPathFromHashList(project_id, id_list)
    data = {}

    # We age going through each samples individually because of a bug
    # (the data aren't aligned with the requested samples id)
    # Because of this bug, we are slowing down the process
    # TODO : fix this bug
    for i in range(len(id_list)):
        # Get tree from samples
        samples_tree = tree.getBlockTreeFromSamples(project_id, [sample_path[i]])

        # Convert tree to array
        data_array = _tree_to_array(samples_tree)

        # Convert array to dict
        data[id_list[i]] = data_array[0]

    return data


def _tree_to_array(tree):
    data_array = []
    for block in tree:
        data_array += _block_to_array_recur(block)
    return data_array


def _get_block_values(block):
    # Adding the block name into the values
    values = [block["name"]]

    # store all key-values into an array
    for data_type in DATA_TYPES:
        if data_type in block:
            for key in range(len(block[data_type])):
                values.append(block[data_type][key])

    return values


def _block_to_array_recur(block):
    # Getting bloc values
    values = _get_block_values(block)
    if "childrenInfoList" not in block or len(block["childrenInfoList"]) == 0:
        return [values]

    else:
        # Getting all child values
        child_values = []
        for child_block in block["childrenInfoList"]:
            child_values += _block_to_array_recur(child_block)
            # child_values.append(_block_to_array_recur(child_block))

        # Child values : [[1,2,3], [4,5,6], [7,8,9]]
        # values : [10, 11, 12]
        # Goal: [[10, 11, 12, 1, 2, 3], [10, 11, 12, 4, 5, 6], [10, 11, 12, 7, 8, 9]]

        # Adding the block name into the values
        ret = [None] * len(child_values)

        # Adding the block values to the children values
        for i in range(len(child_values)):
            ret[i] = values + child_values[i]

        return ret



import requests
import connexion
from flask_cors import CORS
from waitress import serve
from termcolor import colored
from debiaiServer.init import init
from debiaiServer.utils.utils import get_app_version
from debiaiServer.config.init_config import DEBUG_COLOR
from flask import send_from_directory, request, Response


DEV_FRONTEND_URL = "http://localhost:8080/"
app = connexion.App(__name__)
app.add_api("swagger.yaml", strict_validation=True)
app.add_api(
    "exploration/swagger.yaml", strict_validation=True, base_path="/exploration"
)
app.add_api(
    "exploration_statistics/swagger.yaml",
    strict_validation=True,
    base_path="/exploration/statistics",
)

# We duplicate API as V0 API for debiai, it will help to understand refactoring
# - Step 1 duplicate current API as /api/v0 sub API (iso API content)
app.add_api("swagger.yaml", strict_validation=True, base_path="/api/v0")
app.add_api(
    "exploration/swagger.yaml", strict_validation=True, base_path="/api/v0/exploration"
)
app.add_api(
    "exploration_statistics/swagger.yaml",
    strict_validation=True,
    base_path="/api/v0/exploration/statistics",
)


# API Refactoring in /api/V1
# - Step 2 split the main swagger in severals sub API (iso API content)
app.add_api("api/v1/debiai/swagger.yaml", strict_validation=True, base_path="/api/v1")
app.add_api(
    "api/v1/json_block_provider/swagger.yaml",
    strict_validation=True,
    base_path="/api/v1/data",
)
app.add_api(
    "api/v1/exploration/swagger.yaml",
    strict_validation=True,
    base_path="/api/v1/exploration",
)
app.add_api(
    "api/v1/exploration_statistics/swagger.yaml",
    strict_validation=True,
    base_path="/api/v1/exploration/statistics",
)  # noqa: E501

CORS(app.app)


def send_frontend(path, is_dev):
    if path == "/":
        path = "index.html"

    # If production, use the index.html from the dist folder
    if not is_dev:
        return send_from_directory("dist", path)

    # In development, redirect to the DEV_FRONTEND_URL
    else:
        if request.method == "GET":
            try:
                resp = requests.get(f"{DEV_FRONTEND_URL}{path}")
                excluded_headers = [
                    "content-encoding",
                    "content-length",
                    "transfer-encoding",
                    "connection",
                ]
                headers = [
                    (name, value)
                    for (name, value) in resp.raw.headers.items()
                    if name.lower() not in excluded_headers
                ]
                response = Response(resp.content, resp.status_code, headers)
                return response
            except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
                return (
                    "You are in a development environment and the DebiAI frontend"
                    + "is not available at the url : "
                    + DEV_FRONTEND_URL,
                    503,
                )
        else:
            print("Unexpected request method")


def create_app(is_dev):
    # For serving the dashboard
    @app.route("/")
    def send_index():
        return send_frontend("/", is_dev)

    # For serving the dashboard assets
    @app.route("/<path:path>")
    def send_supporting_elements(path):
        return send_frontend(path, is_dev)

    return app


def start_server(
    port,
    data_folder_path: str = None,
    parameters: dict = {},
    reloader: bool = True,
    is_dev: bool = True,
):
    # Run DebiAI init
    print("================= DebiAI " + get_app_version() + " ====================")
    init(data_folder_path, parameters)
    print("======================== RUN =======================")
    print(
        "   DebiAI is available at "
        + colored("http://localhost:" + str(port), DEBUG_COLOR)
    )
    app = create_app(is_dev)

    if is_dev:
        # Use flask server for development
        app.run(port, debug=True, host="0.0.0.0", use_reloader=reloader)
    else:
        # Use waitress for production
        try:
            serve(app, host="0.0.0.0", port=port, threads=6)
        except OSError:
            print(
                colored(
                    f"Port {port} is already in use. Please try another port.", "red"
                )
            )

    # Print goodbye message
    print("\n\n  Goodbye!")
    print("Thank you for using DebiAI.\n")


if __name__ == "__main__":
    start_server(port=3000, reloader=False, is_dev=True)



from debiaiServer.modules.dataProviders.DataProvider import DataProvider
import debiaiServer.modules.dataProviders.dataProviderManager as data_provider_manager
from pickledb import PickleDB
import uuid
from typing import Any, List, Dict

# Create or load a database
project_explorations_db = PickleDB("projectsExplorationsStatistics.db")


# Statistics
def set_column_type(column) -> str:
    # Return mixed if the column has more than one type
    # Except if the column has numbers and text
    nb_types = 0

    if ("hasNumbers" in column["tags"]) or ("hasText" in column["tags"]):
        nb_types += 1
    if "hasList" in column["tags"]:
        nb_types += 1
    if "hasDict" in column["tags"]:
        nb_types += 1
    if nb_types > 1:
        return "mixed"

    # Return the type of the column
    # text overrides number
    if "hasText" in column["tags"]:
        return "text"
    if "hasNumbers" in column["tags"]:
        return "number"
    if "hasList" in column["tags"]:
        return "list"
    if "hasDict" in column["tags"]:
        return "dict"

    return "other"


def can_calculate(column):
    # Check if the column can be calculated
    # Only columns that have numbers or text can be calculated
    if "hasNumbers" in column["tags"] and not (
        "hasText" in column["tags"]
        or "hasOther" in column["tags"]
        or "hasList" in column["tags"]
        or "hasDict" in column["tags"]
    ):
        return True
    return False


def get_columns_statistics(dataProviderId, projectId):
    SAMPLES_PER_PAGE = 50000

    print(f" - Getting columns statistics for project {projectId} on {dataProviderId}")

    # To remove during V0 legacy API removal in future versions
    if dataProviderId == "json_block":
        dataProviderId = "Python module Data Provider"

    # Find the data provider & project
    data_provider = data_provider_manager.get_single_data_provider(dataProviderId)
    data_provider_project = data_provider.get_project(projectId)

    # Check if we already have the statistics
    project_columns_statistics_db_key = f"{dataProviderId}_{projectId}"
    existing = project_explorations_db.get(project_columns_statistics_db_key)
    if existing:
        # Get the project update time
        project_update_time = data_provider_project.get("updateDate", 0)
        if type(existing) is dict:
            existing_update_time = existing.get("updateDate", 0)
            if (existing_update_time or 0) >= (project_update_time or 0):
                return {"columns": existing["columns"]}

    # Get the project columns structure
    print(" - Getting columns structure")
    columns_structure = data_provider_project["columns"]
    # [
    #     {"name": "image", "category": "other", "type": "auto"},
    #     {"name": "objects", "category": "groundtruth", "type": "list"},
    #     {"name": "camera", "category": "context", "type": "text"},
    #     {"name": "objects number", "category": "context", "type": "number"},
    #     {"name": "per classes number", "category": "context", "type": "number"},
    # ]

    # In the structure we have the following elements
    # - name
    # - type
    # - tags (list of tags among which)
    #       "hasNumbers"
    #       "hasText"
    #       "hasNull"
    #       "hasList"
    #       "hasDict"
    #       "hasOther"
    # - metadata (list of information)
    #       "category" : => if not define <=> auto
    # - metrics (list of columns metrics)
    #       - NbUniques
    #       - ...
    #
    # Set the columns statistics
    columns_statistics = []
    metrics_already_present = True
    for column in columns_structure:
        column_statistics = {
            "name": column["name"],
            "type": column["type"],
            "tags": [],
            "metadata": {
                "category": column["category"],
            },
            "metrics": {},
        }

        if column["type"] != "auto":
            if (
                "metrics" not in column
                or type(column["metrics"]) is not dict
                or len(column["metrics"]) == 0
                or "nbUniqueValues" not in column["metrics"]
            ):
                metrics_already_present = False
                column_statistics["metrics"] = {
                    "nbUniqueValues": 0,
                    "nbNullValues": 0,
                    "average": None,
                    "min": None,
                    "max": None,
                }
            else:
                column_statistics["metrics"] = column["metrics"]
        columns_statistics.append(column_statistics)

    # Get the columns data to calculate the statistics if not already present

    if metrics_already_present:
        project_explorations_db.set(
            project_columns_statistics_db_key,
            {
                "dataProviderId": dataProviderId,
                "projectId": projectId,
                "columns": columns_statistics,
                "updateDate": data_provider_project["updateDate"],
            },
        )
        project_explorations_db.save()
        return {"columns": columns_statistics}

    analysis_id = str(uuid.uuid4())
    unique_values_map = {}
    nb_samples = 0

    print(" - Getting samples id list")

    while True:
        # Get the samples id list
        analysis = {"id": analysis_id, "first": nb_samples == 0, "last": False}
        data_id_list = data_provider.get_id_list(
            projectId,
            analysis,
            nb_samples,
            nb_samples + SAMPLES_PER_PAGE - 1,
        )

        # Load the samples
        analysis = {
            "id": analysis_id,
            "first": nb_samples == 0,
            "last": False,
        }

        # Get the samples
        data = data_provider.get_samples(projectId, analysis, data_id_list)

        # Do some statistics on the samples
        current_sample_nb = nb_samples
        for sample_id, sample_values in data.items():
            # For each value in the sample
            for value_index, sample in enumerate(sample_values):
                current_sample_nb += 1
                # Get the column
                column = columns_statistics[value_index]

                # Check if the column value is None
                if sample is None:
                    if "hasNull" not in column["tags"]:
                        column["tags"].append("hasNull")
                    column["metrics"]["nbNullValues"] += 1
                    continue

                # Check the sample type
                if isinstance(sample, (int, float)):
                    if "hasNumbers" not in column["tags"]:
                        column["tags"].append("hasNumbers")
                elif isinstance(sample, str):
                    if "hasText" not in column["tags"]:
                        column["tags"].append("hasText")
                elif isinstance(sample, list):
                    if "hasList" not in column["tags"]:
                        column["tags"].append("hasList")
                elif isinstance(sample, dict):
                    if "hasDict" not in column["tags"]:
                        column["tags"].append("hasDict")
                else:
                    print("Unknown type", type(sample))
                    if "hasOther" not in column["tags"]:
                        column["tags"].append("hasOther")

                # Deal with the unique values
                if "hasNumbers" in column["tags"] or "hasText" in column["tags"]:
                    if column["name"] not in unique_values_map:
                        unique_values_map[column["name"]] = set()
                    unique_values_map[column["name"]].add(sample)

                # Update the min, max and average
                if can_calculate(column):
                    # Update the min, max and average
                    if (
                        column["metrics"]["min"] is None
                        or sample < column["metrics"]["min"]
                    ):
                        column["metrics"]["min"] = sample
                    if (
                        column["metrics"]["max"] is None
                        or sample > column["metrics"]["max"]
                    ):
                        column["metrics"]["max"] = sample
                    if column["metrics"]["average"] is None:
                        column["metrics"]["average"] = sample
                    else:
                        column["metrics"]["average"] = (
                            column["metrics"]["average"] * (current_sample_nb - 1)
                            + sample
                        ) / current_sample_nb

        # If no more samples, break the loop
        nb_samples += len(data_id_list)
        print("  - nb_samples", nb_samples)
        if len(data_id_list) < SAMPLES_PER_PAGE:
            break

    # Final processing of the columns statistics
    for column in columns_statistics:
        # Set the unique values nb
        # Skip the columns that are not of type number or text
        if column["name"] in unique_values_map and not (
            "hasDict" in column["tags"]
            or "hasList" in column["tags"]
            or "hasOther" in column["tags"]
        ):
            column["metrics"]["nbUniqueValues"] = len(unique_values_map[column["name"]])

        # Set the type of the column
        column["type"] = set_column_type(column)

    # Save the statistics in the database
    project_explorations_db.set(
        project_columns_statistics_db_key,
        {
            "dataProviderId": dataProviderId,
            "projectId": projectId,
            "columns": columns_statistics,
            "updateDate": data_provider_project["updateDate"],
        },
    )
    project_explorations_db.save()

    return {"columns": columns_statistics}


# Data loaders
def get_samples_batch(
    data_provider: DataProvider, projectId, analysis, start, end
) -> dict:
    # Get the samples id list
    data_id_list = data_provider.get_id_list(
        projectId,
        analysis,
        start,
        end,
    )

    # Load the samples
    data = data_provider.get_samples(projectId, analysis, data_id_list)

    if "dataMap" in data:
        data = data["data"]

    return data


def get_column_value_in_data(
    column_name: str, sample_data: list, project_columns: list
) -> Any:
    for i, column in enumerate(project_columns):
        if column["name"] == column_name:
            return sample_data[i]
    raise ValueError(f"Column '{column_name}' not found in project columns.")


def get_samples_batch_by_columns(
    data_provider: DataProvider,
    projectId,
    analysis,
    start,
    end,
    project_columns: list,
    selected_columns: list,
    selected_columns_aggregations: dict,
) -> dict:
    """
    Returns a { column_name: [value1, value2, ...] }
    dictionary for the samples in the given range.
    """

    # Get the samples arrays
    data = get_samples_batch(data_provider, projectId, analysis, start, end)

    # Data:
    # {
    #     "sample_id": [
    #         "Value1",
    #         2,
    #         [...],
    #         ...
    #     ],
    #     ...
    # }

    # Get the index of the selected columns in the project columns
    column_indices = {
        column["name"]: i
        for i, column in enumerate(project_columns)
        if column["name"] in selected_columns
    }

    # Convert to a dictionary with column names as keys
    data_columns = {}
    for column_name in selected_columns:
        data_columns[column_name] = [None] * len(data)
        for data_index, sample_data in enumerate(data.values()):
            data_columns[column_name][data_index] = sample_data[
                column_indices[column_name]
            ]

    return data_columns


def get_data_batch(
    data_provider: DataProvider,
    projectId,
    analysis,
    start,
    end,
    project_columns: list,
    selected_columns_labels: List[str],
    columns_aggregation_config: dict,
    columns_statistics: dict,
    metric_columns_to_fetch: list,
) -> Dict[str, dict]:
    """
    Returns a value for each data for the combinations:
    {
        "dataId1": {
            "column_values": [1, "A"],
            "metric_column_values": {
                "col1": 0.5,
                "col2": 1.0,
            },
        },
        "dataId2": {
            "column_values": [2, "B"],
            "metric_column_values": {
                "col1": 0.7,
                "col2": 1.2,
            },
        },
        ...
    }
    """

    # Get the samples arrays
    data = get_samples_batch(data_provider, projectId, analysis, start, end)

    # Data:
    # {
    #     "sample_id": [
    #         "Value1",
    #         2,
    #         [...],
    #         ...
    #     ],
    #     ...
    # }

    # Get the index of the selected columns in the project columns
    column_indices = {
        column["name"]: i
        for i, column in enumerate(project_columns)
        if column["name"] in selected_columns_labels
    }
    metric_column_indices = {
        column["name"]: i
        for i, column in enumerate(project_columns)
        if column["name"] in metric_columns_to_fetch
    }

    # Convert to a dictionary with data id as keys
    data_columns = {}
    for data_id, sample_data in data.items():
        data_columns[data_id] = {
            "column_values": tuple(
                [
                    apply_aggregations(
                        sample_data[column_indices[column_name]],
                        column_name,
                        columns_aggregation_config,
                        columns_statistics,
                    )
                    for column_name in selected_columns_labels
                ]
            ),
            "metric_column_values": {},
        }

        for metric_column_name in metric_columns_to_fetch:
            metric_value = sample_data[metric_column_indices[metric_column_name]]
            data_columns[data_id]["metric_column_values"][
                metric_column_name
            ] = metric_value

    return data_columns


# Aggregations
def apply_aggregations(
    value: Any,
    column_name: str,
    columns_aggregation_config: dict,
    columns_statistics: dict,
) -> Any:
    if not columns_aggregation_config[column_name]:
        return value

    if "type" not in columns_aggregation_config[column_name]:
        print(columns_aggregation_config[column_name])
        raise ValueError(f"Aggregation type not specified for column '{column_name}'")

    aggregation_type = columns_aggregation_config[column_name]["type"]
    if aggregation_type == "evenRange":
        # Get the min and max values from the statistics
        min_value = columns_statistics[column_name]["metrics"]["min"]
        max_value = columns_statistics[column_name]["metrics"]["max"]

        # Get the nb chunks from the aggregation config
        nb_chunks = columns_aggregation_config[column_name]["nbChunks"]

        # Calculate the range size
        range_size = (max_value - min_value) / nb_chunks

        # Calculate the index of the range
        if range_size == 0:
            index = 0
        else:
            index = int((value - min_value) / range_size)
        index = max(0, index)
        index = min(index, nb_chunks - 1)

        # Return "[range_start, range_end]"
        return "[{}, {}]".format(
            round(min_value + index * range_size, 2),
            round(min_value + (index + 1) * range_size, 2),
        )
    if aggregation_type == "firstCharacters":
        # Get the number of characters from the aggregation config
        nb_characters = columns_aggregation_config[column_name]["nbCharacters"]
        nb_characters_skip = columns_aggregation_config[column_name].get(
            "nbCharactersSkip", 0
        )

        # Return the first nb_characters of the value
        return str(value)[
            nb_characters_skip : nb_characters_skip + nb_characters  # noqa: E203
        ]
    else:
        raise ValueError(
            f"Unknown aggregation type '{aggregation_type}' for column '{column_name}'"
        )



from termcolor import colored

from debiaiServer.config.init_config import (
    get_config,
    DEBUG_COLOR,
    ERROR_COLOR,
    SUCCESS_COLOR,
)
from debiaiServer.modules.algoProviders.AlgoProviderException import (
    AlgoProviderException,
)  # noqa
from debiaiServer.modules.algoProviders.AlgoProvider import AlgoProvider
from debiaiServer.modules.algoProviders.integratedAlgoProvider.integratedAlgoProvider import (  # noqa
    IntegratedAlgoProvider,
)

algo_providers = []


def setup_algo_providers():
    print("================== ALGO PROVIDERS ==================")
    config = get_config()
    config_algo_providers = config["ALGO_PROVIDERS"]

    keys = list(config_algo_providers.keys())
    values = list(config_algo_providers.values())

    # Add AlgoProviders from config file
    print(" - Loading Algo providers from config file")
    for i in range(len(config_algo_providers)):
        name = keys[i]
        url = values[i]

        # Remove trailing slash
        if url[-1] == "/":
            url = url[:-1]

        print(
            " - Adding AlgoProvider "
            + colored(name, DEBUG_COLOR)
            + " ("
            + colored(url, DEBUG_COLOR)
            + ")"
        )
        try:
            algo_provider = AlgoProvider(url, name)
            algo_providers.append(algo_provider)

            if algo_provider.is_alive():
                print(
                    colored("   [SUCCESS]", SUCCESS_COLOR)
                    + " AlgoProvider "
                    + colored(name, SUCCESS_COLOR)
                    + " ready"
                )
            else:
                raise AlgoProviderException()

        except AlgoProviderException:
            print(
                colored("   [ERROR]", ERROR_COLOR)
                + " AlgoProvider "
                + colored(name, ERROR_COLOR)
                + " is not accessible"
            )

    # Add the integrated algo provider
    enable_integrated = config["ALGO_PROVIDERS_CONFIG"]["enable_integrated"]
    if enable_integrated:
        print(" - Adding integrated AlgoProviders")
        algo_provider = IntegratedAlgoProvider()
        nb_algos = len(algo_provider.get_algorithms())
        algo_providers.append(algo_provider)

        if nb_algos > 0:
            print(
                colored("   [SUCCESS]", SUCCESS_COLOR)
                + " Integrated AlgoProvider ready with "
                + str(nb_algos)
                + " algorithms"
            )
        else:
            print("   No algorithms found")

    if len(algo_providers) == 0:
        print("No Algo providers")


def get_algo_providers():
    return algo_providers


def get_algo_providers_json():
    algo_providers_json = []
    for algo_provider in algo_providers:
        algo_providers_json.append(algo_provider.to_json())

    return algo_providers_json


def algo_provider_exists(name):
    for d in algo_providers:
        if d.name == name:
            return True
    return False


def add(algo_provider):
    algo_providers.append(algo_provider)
    return


def get_single_algo_provider(name):
    # Return the algo provider with the given name
    for d in algo_providers:
        if d.name == name:
            return d

    raise AlgoProviderException("Algo provider not found", 404)


def delete(name):
    for d in algo_providers:
        if d.name == name:
            algo_providers.remove(d)
            return



from setuptools import setup, find_packages
from debiaiServer.utils.utils import get_app_version

try:
    VERSION = get_app_version()
except ModuleNotFoundError:
    VERSION = "0.0.0"

setup(
    name="debiai_gui",
    version=VERSION,
    packages=find_packages(include=["debiaiServer", "debiaiServer.*"]),
    include_package_data=True,
    install_requires=[
        "Flask==2.0.3",
        "flask_cors==3.0.8",
        "connexion==2.6.0",
        "requests==2.25.1",
        "swagger-ui-bundle==0.0.5",
        "ujson==5.8.0",
        "kafka-python==2.0.2",
        "openapi_spec_validator==0.2.8",
        "PyYAML==6.0",
        "cacheout==0.14.1",
        "termcolor==2.3.0",
        "werkzeug==2.2.2",
        "psutil==6.0.0",
        "waitress==3.0.0",
        "pickledb==1.3.2",
        'setuptools==80.9.0'
    ],
    entry_points={
        "console_scripts": [
            "debiai-gui=debiaiServer.debiai_gui_utils:main",
        ],
    },
    author="IRT-Systemx",
    author_email="debiai@irt-systemx.fr",
    description="DebiAI easy start module, the standalone version of DebiAI",
    long_description=open("debiai_gui.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/debiai/DebiAI",
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.6",
)



#############################################################################
# Imports
#############################################################################
from debiaiServer.config.init_config import get_config
import debiaiServer.modules.dataProviders.dataProviderManager as data_provider_manager
from debiaiServer.modules.dataProviders.DataProviderException import (
    DataProviderException,
)
from debiaiServer.api.v1.debiai.utils import make_hash


def format_data_provider_info(data_provider):
    data = {
        "id": data_provider.id,
        "status": True,
        "name": data_provider.name,
        "type": data_provider.type,
        "tags": [],
        "metadata": {},
        "metrics": {},
    }
    print("NAME ", data_provider.name)

    provider_info = data_provider.get_info()
    if "canDelete" in provider_info and provider_info["canDelete"]["models"]:
        data["tags"].append("canDeleteModels")
    if "canDelete" in provider_info and provider_info["canDelete"]["projects"]:
        data["tags"].append("canDeleteProject")
    if "canDelete" in provider_info and provider_info["canDelete"]["selections"]:
        data["tags"].append("canDeleteSelections")
    if "version" in provider_info:
        data["metadata"]["version"] = provider_info["version"]

    if "maxResultByRequest" in provider_info:
        data["metrics"]["maxResultByRequest"] = provider_info["maxResultByRequest"]
    if "maxSampleDataByRequest" in provider_info:
        data["metrics"]["maxSampleDataByRequest"] = provider_info[
            "maxSampleDataByRequest"
        ]
    if "maxSampleIdByRequest" in provider_info:
        data["metrics"]["maxSampleIdByRequest"] = provider_info["maxSampleIdByRequest"]

    if data_provider.name != "Python module Data Provider":
        data["metadata"]["external_url"] = data_provider.url

    return data


def get_data_provider_info(dataProviderId):

    # As Parquet not yet supported we force json_block:
    if dataProviderId == "json_block":
        dataProviderId = "Python module Data Provider"

    try:
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)
        data = format_data_provider_info(data_provider)

        return data, 200
    except DataProviderException as e:
        return e.message, e.status_code


def get_data_providers(prev_hash_content=None):
    data_provider_list = data_provider_manager.get_data_provider_list()
    providers_formatted = []

    for data_provider in data_provider_list:
        try:
            data = format_data_provider_info(data_provider)
        except DataProviderException as e:
            data = {
                "id": data_provider.id,
                "status": False,
                "name": data_provider.name,
                "type": data_provider.type,
                "tags": [],
                "metadata": {
                    "status_code": e.status_code,
                    "message": e.message,
                },
                "metrics": {},
            }

        providers_formatted.append(data)

    new_hash = "data_" + str(make_hash(providers_formatted))
    # We add a prefix to avoir empty string

    print(
        new_hash,
        " <=> ",
        prev_hash_content,
        type(new_hash),
        type(prev_hash_content),
        new_hash == prev_hash_content,
    )
    if new_hash == prev_hash_content:
        return None, 304
    else:
        providers_answer = {
            "dataproviders": providers_formatted,
            "hash_content": new_hash,
        }
        return providers_answer, 200


def delete_data_providers(dataProviderId):
    # Check if we are allowed to add data providers from the config file
    config = get_config()
    deletion_allowed = config["DATA_PROVIDERS_CONFIG"]["deletion"]
    if not deletion_allowed:
        return "Data provider deletion is not allowed", 403

    # To remove when legacy API V0 removal
    if dataProviderId == "json_block":
        dataProviderId = "Python module Data Provider"

    # Delete data provider
    try:
        data_provider_manager.delete(dataProviderId)
        return None, 204
    except DataProviderException as e:
        return e.message, e.status_code



import uuid
import threading
from typing import List
from .utils import (
    explorations_db,
    start_exploration_real_combination_computation,
    create_selection,
    update_exploration_db,
    computation_threads,
    stop_flags,
)
from debiaiServer.api.v1.debiai.utils import make_hash


def get_exploration_available_config():
    pass


def get_explorations(project_id, prev_hash_content=None):
    explorations: List[dict] = []
    for exploration_id in explorations_db.all():
        exploration = explorations_db.get(exploration_id)
        if exploration.get("project_id") != project_id:
            continue

        exploration.pop("combinations", None)
        exploration.pop("metrics", None)
        explorations.append(exploration)

    new_hash = "exploration_" + str(make_hash(explorations))

    if new_hash == prev_hash_content:
        return None, 304
    else:
        explorations_answer = {
            "explorations": explorations,
            "hash_content": new_hash,
        }
        return explorations_answer, 200


def create_exploration(project_id, body):
    # Get current explorations
    exploration_id = str(uuid.uuid4())
    new_exploration = {
        "id": exploration_id,
        "project_id": project_id,
        "name": body.get("name"),
        "description": body.get("description", ""),
        "state": "not_started",
        "config": {},
    }

    # Add new exploration
    explorations_db.set(exploration_id, new_exploration)
    explorations_db.save()
    return 201


def get_exploration(exploration_id):
    exploration = explorations_db.get(exploration_id)
    if exploration:
        return exploration, 200
    else:
        return 404


def delete_exploration(exploration_id):
    explorations_db.remove(exploration_id)
    return 204


def update_exploration(exploration_id, action, body):
    exploration = explorations_db.get(exploration_id)
    if exploration is None:
        print("Exploration not found:", exploration_id)
        return None, 404

    # Perform action based on the request
    if action == "start":
        # Check if the exploration is already running
        if exploration.get("state") == "ongoing":
            print("Exploration is already running:", exploration_id)
            return exploration

        # Start the exploration computation in a separate thread
        print(exploration_id)
        thread = threading.Thread(
            target=start_exploration_real_combination_computation,
            args=(exploration_id,),
            daemon=True,
        )
        thread.start()

        exploration["state"] = "ongoing"
        # Save the thread in the global dictionary
        computation_threads[exploration_id] = thread
        update_exploration_db(exploration_id, exploration)

        return exploration
    elif action == "stop":
        # Stop the exploration computation
        if exploration_id in computation_threads:
            # Set the stop flag
            stop_flags[exploration_id] = True

            # Wait for the thread to finish
            computation_threads[exploration_id].join()
            del computation_threads[exploration_id]
            del stop_flags[exploration_id]
        elif exploration.get("state") == "ongoing":
            # Exploration is ongoing but no thread found, set state to not_started
            exploration["state"] = "not_started"

        update_exploration_db(exploration_id, exploration)
        return exploration
    elif action == "updateConfig":
        # Do not update the exploration if it is running
        if exploration.get("state") == "ongoing":
            print("Cannot update config while exploration is running:", exploration_id)
            return exploration

        # Update exploration config
        exploration["config"] = body
        update_exploration_db(exploration_id, exploration)
        return exploration
    else:
        raise ValueError(f"Unknown action: {action}")


def update_exploration_config():
    pass


def create_exploration_selection(exploration_id, body):
    exploration = explorations_db.get(exploration_id)
    # Create a selection for the exploration
    create_selection(
        exploration.get("project_id"),
        exploration_id,
        body["selected_combinations"],
        body["selection_name"],
    )

    return 201



#############################################################################
# Imports
#############################################################################
import debiaiServer.utils.layouts.layouts as layoutsUtils
from debiaiServer.api.v1.debiai.utils import make_hash

#############################################################################
# Analysis dashboard layout Management
#############################################################################


def get_layouts(prev_hash_content=None):
    layouts_overview = layoutsUtils.get_layouts()
    for layout in layouts_overview:
        layout.pop("dataProviderId", None)
        print(layout)

    new_hash = "layout_" + str(make_hash(layouts_overview))
    print(new_hash, " <=> ", prev_hash_content, new_hash == prev_hash_content)

    if new_hash == prev_hash_content:
        return None, 304
    else:
        layouts_answer = {
            "layouts": layouts_overview,
            "hash_content": new_hash,
        }
        return layouts_answer, 200


def post_layout(data):
    layoutsUtils.add_layout(data)
    return None, 204


def delete_layout(id):
    layoutsUtils.delete_layout(id)
    return None, 204
